#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <unistd.h>

int bail(const char *context) {
  fprintf(stderr, "%s: %s\n", context, strerror(errno));
  return EXIT_FAILURE;
}

int main() {
  const char *host = "info.cern.ch";
  int sock = -1;

  int rv;                  // Store the return value
  struct addrinfo *result; // Pointer to where we store the addresses
  struct addrinfo *rp;     // Later used to connect to address
  struct addrinfo hints;   // Indicates, which service address we want to get

  // Set the entire struct to 0
  memset(&hints, 0, sizeof(hints));

  // Set hints to allow for IPv4 address and for a TCP connection
  hints.ai_family = AF_INET;
  hints.ai_socktype = SOCK_STREAM;

  rv = getaddrinfo(host, "http", &hints, &result);
  if (rv != 0) {
    fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(rv));
    return bail("Could not find address of remote service");
  }

  // Since getaddrinfo() returns a list of address structures
  // Try each address until we successfully connect()
  // If socket() (or connect()) fails, we (close the socket and)
  // try the next address
  for (rp = result; rp != NULL; rp = rp->ai_next) {
    sock = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
    if (sock == -1) {
      continue;
    }

    if (connect(sock, rp->ai_addr, rp->ai_addrlen) != -1) {
      break;
    }

    close(sock);
  }

  // No longer needed
  freeaddrinfo(result);

  if (rp == NULL) {
    return bail("Could not connect to remote server");
  }
  // Building the request and getting its length
  const char *request =
      "GET / HTTP/1.1\r\n"
      "Host: info.cern.ch\r\n"
      "Accept-Encoding: identity\r\n"
      "\r\n"; // The last \r\n is important! Otherwise HTTP expects more data
  int requestLen = strlen(request);

  // Making sure, all the bytes from the request are being sent
  while (requestLen > 0) {
    int n = send(sock, request, requestLen, 0);
    if (n < 0) {
      return bail("The amount of sent bytes is negative");
    }
    requestLen -= n;
    request += n;
  }
  // Amount of outstanding requests, important because we multiplex multiple
  // connections into a file descriptor set
  int outstanding = 1;

  while (outstanding) {
    fd_set fds;
    FD_ZERO(&fds);
    FD_SET(sock, &fds);

    struct timeval tv;
    tv.tv_sec = 0;
    tv.tv_usec = 0;

    int r = select(sock + 1, &fds, NULL, NULL, &tv);
    if (r < 0) {
      return bail("File descriptor set is has invalid state");
    }

    if (FD_ISSET(sock, &fds) == 0) {
      // No data needs to be read, so we continue in the loop
      continue;
    }

    unsigned char buf[2048];
    int a = recv(sock, (char *)buf, sizeof(buf), 0);
    if (a < 0) {
      return bail("Read a negative amount of bytes from socket");
    }
    // In this case, data needs to be read from the socket
    if (a != 0) {
      int used = 0;
      while (used < a) {
        // I'm feeling lazy, so I just print the data
        printf("%s", buf);
        used += a - used;
      }

      // In this case, all the data from the socket was read
    } else {
      // For multiple connection, you'd probably want this to decrement instead
      // of setting it to 0
      outstanding = 0;

      // Closing socket
      close(sock);
      sock = -1;
    }
  }

  return EXIT_SUCCESS;
}
